"""
Manage your Anaconda Team Edition server-side settings

###### Set server-side settings

You can add a site named **site_name** like this:

    conda repo admin --set SETTING_NAME SETTING_VALUE

Currently supported settings are:

    - `user_channel_autocreate` - if True, a user's channel is created automatically on first login

###### Show server-side settings

Yuo can see current settings:

    conda repo admin --show
"""
from argparse import RawDescriptionHelpFormatter
import logging

from .base import SubCommandBase
from .. import errors
from ..utils.format import SettingsFormatter
from ..utils.yaml import safe_load


logger = logging.getLogger('repo_cli')


SETTINGS_VALIDATOR = {
    'user_channel_autocreate': safe_load
}


class SubCommand(SubCommandBase):
    name = 'admin'

    def main(self):

        if self.args.show:
            self.show_settings()
            return

        if self.args.get:
            self.show_settings(self.args.get)
            return

        if self.args.set:
            self.update_settings(self.args.set)
            return

        raise NotImplementedError('Please use command options')

    def show_settings(self, key=None):
        settings = self.api.get_system_settings()
        if key is not None:
            if key not in settings:
                raise errors.RepoCLIError('%s is an unknown admin setting' % key)
            settings = {
                key: settings[key]
            }

        self.log.info(SettingsFormatter.format_object_as_list(settings))

    def update_settings(self, args):
        data = {}
        for key, value in args:
            if key not in SETTINGS_VALIDATOR:
                raise errors.RepoCLIError('%s is an unknown admin setting' % key)
            data[key] = SETTINGS_VALIDATOR[key](value)

        settings = self.api.get_system_settings()
        settings.update(data)
        self.api.update_system_settings(settings)

        self.log.info('Anaconda Team Edition settings are updated')

    def add_parser(self, subparsers):
        description = 'Anaconda Team Edition admin settings'
        parser = subparsers.add_parser('admin',
                                       help=description,
                                       description=description,
                                       epilog=__doc__,
                                       formatter_class=RawDescriptionHelpFormatter)

        agroup = parser.add_argument_group('actions')

        agroup.add_argument('--set', nargs=2, action='append', default=[],
                            help='sets a server setting value: name value', metavar=('name', 'value'))
        agroup.add_argument('--get', metavar='name',
                            help='get value: name')
        agroup.add_argument('--show', action='store_true', default=False,
                            help='show all variables')

        parser.set_defaults(main=self.main, sub_parser=parser)



import pytest
import responses
from unittest.mock import patch

from repo_cli import main
from repo_cli.utils import config
from repo_cli.tests.conftest import default_username, run_patched
from .testcase import load_mocks

default_config = config.get_config()
test_site = default_config.get("default_site", "test")

url = default_config.get('url', "http://localhost:8088/api/")


def test_login(default_user):
    main.main(["logout"])
    token = config.load_token(test_site)
    assert not token

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        load_mocks(rsps)

        main.main(
            [
                "login",
                "--username=%s" % default_user["username"],
                "--password=%s" % default_user["password"],
            ]
        )
        token = config.load_token(test_site)
        assert token


def test_logout(token_user_1):
    main.main(["logout"])
    token = config.load_token(test_site)
    assert not token


def test_list_auth_scopes(caplog):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        load_mocks(rsps)

        main.main(["auth", "--list-scopes"])

        assert (
            "artifact:create,artifact:delete,artifact:download,artifact:edit,artifact:view,"
            "channel.default-channel:edit,channel.group:edit,channel:create,channel:delete,"
            "channel:edit,channel:history,channel:view,channel:view-artifacts,subchannel.group:edit,"
            "subchannel:create,subchannel:delete,subchannel:edit,subchannel:history,subchannel:view,"
            "subchannel:view-artifacts" in caplog.text
        ), caplog.text


# @pytest.mark.skip('CBR-1694')
def test_list_user_tokens(caplog, monkeypatch, token_user_1):
    monkeypatch.setattr(main.RepoCommand, "run", run_patched)

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        load_mocks(rsps)

        main.main(["auth", "--list"])

        assert "repo-cli-token" in caplog.text, caplog.text


# @pytest.mark.skip("auth subcommand logger will cause failure")
def test_create_user_token(caplog, monkeypatch, token_user_1):
    monkeypatch.setattr(main.RepoCommand, "run", run_patched)
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        load_mocks(rsps)
        main.main(["auth", "--create", "-n", "john", "-s", "artifact:view", "-s", "cve:view"])
        assert "Token 18ba3a4f7d8681adb5d0af712d14d77f2db5cccd6eb0cbc7 " \
               "succesfully created with id: 6b39b019-8e28-4fa1-8249-7055cf36b9ef" in caplog.text, caplog.text


@pytest.mark.skip("auth subcommand logger will cuase failure")
@patch("builtins.input", lambda *args: default_username)
def test_remove_user_token(token_user_1):
    main.main(["-t", token_user_1, "auth", "--remove", token_user_1])
    token = config.load_token(test_site)
    assert not token



'''
Remove an object from your Anaconda Team Edition.

Example::


    # Remove specific package by providing version and package filename
    conda repo remove mychannel::mypackage/1.2.0/mypackage.tar.gz

    # Remove specific package by providing package filename
    conda repo remove mychannel::mypackage//mypackage.tar.gz

    # Remove all package files
    conda repo remove mychannel::mypackage

'''
from six.moves.urllib.parse import urlparse
import sys
from argparse import RawTextHelpFormatter
from .. import errors
from .base import SubCommandBase


def bool_input(prompt, default=True):
    default_str = '[Y|n]' if default else '[y|N]'
    while 1:
        inpt = input('%s %s: ' % (prompt, default_str))
        if inpt.lower() in ['y', 'yes'] and not default:
            return True
        elif inpt.lower() in ['', 'n', 'no'] and not default:
            return False
        elif inpt.lower() in ['', 'y', 'yes']:
            return True
        elif inpt.lower() in ['n', 'no']:
            return False
        else:
            sys.stderr.write('please enter yes or no\n')


WAIT_SECONDS = 15


class PackageSpec(object):
    def __init__(self, channel, package=None, version=None, filename=None, attrs=None, spec_str=None):
        self._user = channel
        self._package = package
        self._version = version
        self._filename = filename
        self.attrs = attrs
        if spec_str:
            self.spec_str = spec_str
        else:
            spec_str = str(channel)
            if package:
                spec_str = '%s/%s' % (spec_str, package)
            if version:
                spec_str = '%s/%s' % (spec_str, version)
            if filename:
                spec_str = '%s/%s' % (spec_str, filename)
            self.spec_str = spec_str

    @classmethod
    def from_string(cls, spec_string):
        channel = spec_string
        package = version = filename = None
        attrs = {}
        if '::' in channel:
            channel, package = channel.split('::', 1)
        if package and '/' in package:
            package, version = package.split('/', 1)

        if version and '/' in version:
            version, filename = version.split('/', 1)

        if filename and '?' in filename:
            filename, qsl = filename.rsplit('?', 1)
            attrs = dict(urlparse.parse_qsl(qsl))

        return PackageSpec(channel, package, version, filename, attrs, spec_string)

    def __str__(self):
        return self.spec_str

    def __repr__(self):
        return '<PackageSpec %r>' % (self.spec_str)

    @property
    def channel(self):
        if self._user is None:
            raise errors.UserError('user not given (got %r expected <username> )' % (self.spec_str,))
        return self._user

    @property
    def name(self):
        if self._package is None:
            raise errors.UserError('package not given in spec (got %r expected <username>/<package> )' % (self.spec_str,))
        return self._package

    @property
    def package(self):
        if self._package is None:
            raise errors.UserError('package not given in spec (got %r expected <username>/<package> )' % (self.spec_str,))
        return self._package

    @property
    def version(self):
        if self._version is None:
            raise errors.UserError('version not given in spec (got %r expected <username>/<package>/<version> )' % (self.spec_str,))
        return self._version

    @property
    def filename(self):
        if self._filename is None:
            raise errors.UserError('filename not given in spec (got %r expected <username>/<package>/<version>/<filename> )' % (self.spec_str,))
        return self._filename

class SubCommand(SubCommandBase):
    name = "remove"

    def main(self):

        for spec in self.args.specs:
            try:
                if spec._filename:
                    self.remove_artifact(spec.channel, self.args.family, spec.package, spec.version, spec.filename, spec)
                elif spec._version:
                    self.remove_artifact(spec.channel, self.args.family, spec.package, spec.version, spec=spec)
                elif spec._package:
                    self.remove_artifact(spec.channel, self.args.family, spec.package, spec=spec)
                else:
                    self.log.error('Invalid package specification: %s', spec)

            except errors.NotFound:
                if self.args.force:
                    self.log.warning('', exc_info=True)
                    continue
                else:
                    raise

    def remove_artifact(self, channel, family, artifact, version=None, filename=None, spec=None):
        base_item = {
                "name": artifact,
                "family": family,
        }

        items = []
        if version or filename:
            packages, total_count = self.api.get_channel_artifacts_files(channel, family, artifact, version, filename)

            if not packages:
                self.log.warning('No files matches were found for the provided spec: %s\n' % (spec))
                return

            files_descr = []
            for filep in packages:
                files_descr.append('PACKAGE: {name}:{version}-{ckey}; PLATFORM: {platform}; FILENAME: {fn}'.format(**filep))
                item = dict(base_item)
                item['ckey'] = filep['ckey']
                items.append(item)

            affected_files = '\n'.join(files_descr)
            msg = 'Are you sure you want to remove the package release %s ? The following ' \
                  'will be affected: \n\n %s\n\nConfirm?' % (spec, affected_files)

        else:
            msg = 'Are you sure you want to remove the package %s ? (and all data with it?)' % (spec,)
            items = [base_item]

        if self.args.force or bool_input(msg, False):
            self.api.channel_artifacts_bulk_actions(channel, 'delete', items)

            self.log.info('Spec %s succesfully removed\n' % (spec))
        else:
            self.log.warning('Not removing release %s\n' % (spec))


    def add_parser(self, subparsers):

        parser = subparsers.add_parser('remove',
                                          help='Remove an object from your Anaconda Team Edition repository. Must refer to the '
                                               'formal package name as it appears in the URL of the package. Also '
                                               'use anaconda show <USERNAME> to see list of package names. '
                                               'Example: anaconda remove continuumio/empty-example-notebook',
                                          description=__doc__, formatter_class=RawTextHelpFormatter)

        parser.add_argument('specs', help='Package written as <channel>/<subchannel>[::<package>[/<version>[/<filename>]]]',
                            type=PackageSpec.from_string, nargs='+')
        parser.add_argument('-f', '--force', help='Do not prompt removal', action='store_true')
        parser.add_argument('--family', default='conda', help='artifact family (i.e.: conda, python, cran, anaconda_project, '
                                                              'anaconda_env, nootebook)')

        parser.set_defaults(main=self.main)



'''
`conda repo` configuration

Get, Set, Remove or Show the `conda repo` configuration.

###### `conda repo` sites

`conda repo` sites are a mechanism to allow users to quickly switch
between Anaconda repository instances, from anaconda.org to on-site
[Anaconda Repo](http://continuum.io/anaconda-server) installations.

`conda repo` comes one two pre-configured site: `anaconda`. If you
have configured more than one site, you can operate between them
in 2 ways:

  * Invoke the `conda repo` command with the `-s/--site` option
    e.g. to use the alpha testing site:

        conda repo -s mysite whoami

  * Set a site as the default:

        conda repo config --set default_site mysite
        conda repo whoami

  * Show sources:
        conda repo config --show-sources

###### Add a `conda repo` site

You can add a site named **site_name** like this:

    conda repo config --set sites.site_name.url "http://url_of_new_site/api"
    conda repo config --set default_site site_name

###### Site Options VS Global Options

All options can be set as global options - affecting all sites,
or site options - affecting only one site

By default options are set globally e.g.:

    conda repo config --set OPTION VALUE

If you want the option to be limited to a single site,
prefix the option with `sites.site_name` e.g.

    conda repo config --set sites.site_name.OPTION VALUE

###### Common `conda repo` configuration options

  * `url`: Set the conda repo api url (default: https://api.anaconda.org)
  * `ssl_verify`: Perform ssl validation on the https requests.
    ssl_verify may be `True`, `False` or a path to a root CA pem file.


###### Other `conda repo` configuration options
  * `auto_register`: Toggle auto_register when doing conda repo upload
  * 'default_site': Default site set to be used by default
  * 'sites': Sites namespace that can be used to add new sites (for
    more information, check the dedicated section above.


###### Toggle auto_register when doing conda repo upload

The default is yes, automatically create a new package when uploading.
If no, then an upload will fail if the package name does not already exist on the server.

    conda repo config --set auto_register yes|no

'''
from __future__ import print_function

from argparse import RawDescriptionHelpFormatter

from .. import errors
from ..utils.config import (SEARCH_PATH, USER_CONFIG, SYSTEM_CONFIG, CONFIGURATION_KEYS,
                            get_config, save_config, load_config, load_file_configs)
from ..utils.validators import check_url
from ..utils.yaml import yaml_dump, safe_load

from .base import SubCommandBase


DEPRECATED = {}


class SubCommand(SubCommandBase):
    name = "config"

    def main(self):

        config = get_config()

        if self.args.show:
            self.log.info(yaml_dump(config))
            return

        if self.args.show_sources:
            config_files = load_file_configs(SEARCH_PATH)
            for path in config_files:
                self.log.info('==> %s <==', path)
                self.log.info(yaml_dump(config_files[path]))
            return

        if self.args.get:
            if self.args.get in config:
                self.log.info(config[self.args.get])
            else:
                self.log.info("The value of '%s' is not set." % self.args.get)
            return

        if self.args.files:
            self.log.info('User Config: %s' % USER_CONFIG)
            self.log.info('System Config: %s' % SYSTEM_CONFIG)
            return

        config_file = USER_CONFIG if self.args.user else SYSTEM_CONFIG

        config = load_config(config_file)
        config_system = load_config(SYSTEM_CONFIG)

        for key, value in self.args.set:
            self._validate_url_key(key, value)
            if self.args.system:
                self.recursive_set(config_system, key, value, self.args.type)    
            self.recursive_set(config, key, value, self.args.type)

        for key in self.args.remove:
            try:
                self.recursive_remove(config, key)
            except KeyError:
                self.log.error("Key %s does not exist" % key)

        if not (self.args.set or self.args.remove):
            raise errors.ShowHelp()

        save_config(config, config_file)
        if self.args.system:
            save_config(config_system, SYSTEM_CONFIG)
    def recursive_remove(self, config_data, key):
        while '.' in key:
            if not config_data:
                return
            prefix, key = key.split('.', 1)
            config_data = config_data.get(prefix, {})

        del config_data[key]

    def recursive_set(self, config_data, key, value, type_):
        while '.' in key:
            prefix, key = key.split('.', 1)
            config_data = config_data.setdefault(prefix, {})

        if key not in CONFIGURATION_KEYS:
            self.log.warning('"%s" is not a known configuration key', key)

        if key in DEPRECATED.keys():
            message = "{} is deprecated: {}".format(key, DEPRECATED[key])
            self.log.warning(message)

        config_data[key] = type_(value)

    def _validate_url_key(self, key, value):
        if key.endswith(".url"):
            check_url((value))

    def add_parser(self, subparsers):
        description = 'Anaconda Repo Client Configuration'
        parser = subparsers.add_parser('config',
                                       help=description,
                                       description=description,
                                       epilog=__doc__,
                                       formatter_class=RawDescriptionHelpFormatter)

        parser.add_argument('--type', default=safe_load,
                            help='The type of the values in the set commands. NOTE: This'
                                 'argument is only used when combined with the `--set`'
                                 'command and is ignored otherwise.')

        agroup = parser.add_argument_group('actions')

        agroup.add_argument('--set', nargs=2, action='append', default=[],
                            help='sets a new variable: name value', metavar=('name', 'value'))
        agroup.add_argument('--get', metavar='name',
                            help='get value: name')
        agroup.add_argument('--remove', action='append', default=[],
                            help='removes a variable')
        agroup.add_argument('--show', action='store_true', default=False,
                            help='show all variables')
        agroup.add_argument('-f', '--files', action='store_true',
                            help='show the config file names')
        agroup.add_argument('--show-sources', action='store_true',
                            help='Display all identified config sources')
        lgroup = parser.add_argument_group('location')
        lgroup.add_argument('-u', '--user', action='store_true', dest='user', default=True,
                            help='set a variable for this user')
        lgroup.add_argument('-s', '--system', '--site', action='store_true', dest='system', default=False,
                            help='set a variable for all users on this machine')

        parser.set_defaults(main=self.main, sub_parser=parser)



"""
Authenticate a user
"""
from __future__ import unicode_literals
import json
import getpass
from os.path import join
import logging
import platform
import sys
import re
import requests
from six.moves import input
from urllib.parse import urlparse
import webbrowser

from ..utils.config import (
    store_token, get_config, OIDC_CLIENT_ID
)
from .. import errors
from ..utils.api import RepoApi
from ..utils.local_auth_server import WebServer
from .base import SubCommandBase

logger = logging.getLogger('repo_cli')


def get_login_and_password(args):
    if getattr(args, 'login_username', None):
        username = args.login_username
    else:
        username = input('Username: ')
    #
    password = getattr(args, 'login_password', None)
    return username, password


def interactive_get_token(args, fail_if_already_exists=True):
    # bs = get_server_api(args.token, args.site)
    config = get_config(site=args.site)

    token = None
    # This function could be called from a totally different CLI, so we don't
    # know if the attribute hostname exists.
    # args.site or config.get('default_site')
    url = config.get('url')
    username, password = get_login_and_password(args)

    for _ in range(3):
        try:
            if password is None:
                password = getpass.getpass(stream=sys.stderr)

            api = RepoApi(base_url=url)
            token = api.login(username, password)

            # token = login_user(username, password, url)
            if not token:
                msg = 'Unable to request the user token. Server was unable to return any valid token!'
                logger.error(msg)
                raise errors.RepoCLIError(msg)
            return token['user']

        except errors.Unauthorized:
            logger.error('Invalid Username password combination, please try again')
            password = None
            continue

    return token['user']


def interactive_login(args):
    token = interactive_get_token(args)
    store_token(token, args)
    logger.info('login successful')
    return token


def main(args):
    interactive_login(args)


class SubCommand(SubCommandBase):
    name = "login"
    manages_auth = True

    def main(self):
        self.login()

    def login(self):
        token = self.interactive_get_token()
        store_token(token, self.args)
        is_admin = self.api.is_admin_jwt
        msg = 'login as ADMIN successful' if is_admin else 'login successful'
        self.log.info(msg)
        return token  # ['id']

    def get_login_and_password(self):
        if getattr(self.args, 'login_username', None):
            username = self.args.login_username
        else:
            username = input('Username: ')
        self.username = username
        password = getattr(self.args, 'login_password', None)
        return username, password

    def interactive_get_token(self):
        config = get_config()
        if config.get('oauth2'):
            token = self.oauth2_get_token()
        else:
            token = self.direct_get_token()
        return token

    def get_and_validate_user_token(self):
        user_token = self.api.get_user_token()

        if not user_token:
            msg = 'Unable to request the user token. Server was unable to return any valid user token!'
            logger.error(msg)
            raise errors.RepoCLIError(msg)

        return user_token

    def direct_get_token(self):
        username, password = self.get_login_and_password()
        for _ in range(3):
            try:
                if password is None:
                    password = getpass.getpass(stream=sys.stderr)

                jwt_token = self.api.login(username, password)
                user_token = self.get_and_validate_user_token()
                return user_token

            except errors.Unauthorized:
                logger.error('Invalid Username password combination, please try again')
                password = None
                continue

        raise errors.RepoCLIError("You've reached maximum login attempts")

    def get_openid_configuration_url(self):
        url = self.api.get_authorize_url()
        matches = re.match('(/auth/realms/([a-z]+)/).*', url.path)
        if not matches:
            logger.info(f'Auth path is not supported: {url.path}')
            raise errors.WrongRepoAuthSetup()

        url = url._replace(path=matches[1] + '.well-known/openid-configuration')
        return url.geturl()

    def oauth2_get_token(self):
        openid_configuration_url=self.get_openid_configuration_url()
        self.log.debug(f"OpenID configuration: {openid_configuration_url}")
        server = WebServer(
            client_id=OIDC_CLIENT_ID,
            openid_configuration_url=openid_configuration_url
        )
        thread = server.start()
        thread.start()
        self.log.info("Opening browser to get a token...")
        webbrowser.open(server.localhost_url())
        thread.join(120)
        if not server.access_token:
            msg = 'Unable to request the user token. Server was unable to return any valid token!'
            logger.error(msg)
            raise errors.RepoCLIError(msg)
        self.api._jwt = server.access_token
        user_token = self.get_and_validate_user_token()
        return user_token

    def add_parser(self, subparsers):
        self.subparser = subparser = subparsers.add_parser('login', help='Authenticate a user', description=__doc__)
        subparser.add_argument('--hostname', default=platform.node(),
                               help="Specify the host name of this login, this should be unique (default: %(default)s)")
        subparser.add_argument('--username', dest='login_username',
                               help="Specify your username. If this is not given, you will be prompted")
        subparser.add_argument('--password', dest='login_password',
                               help="Specify your password. If this is not given, you will be prompted")
        subparser.set_defaults(main=self.main)


from __future__ import unicode_literals

import base64
import datetime
import json
import logging
import os
import re
import socket
from os.path import basename
from posixpath import join
from urllib.parse import urlparse

import requests
import urllib3

from .. import errors
from ..errors import InvalidName, NoDefaultUrl
from ..utils.config import UPLOAD_TYPE_MAPPING

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

logger = logging.getLogger('repo_cli')


ROLE_AUTHOR = 'author'
ROLE_ADMIN = 'admin'


class RepoApi:
    def __init__(self, base_url, user_token=None, verify_ssl=True):
        self.base_url = base_url
        self.verify_ssl = verify_ssl
        self._access_token = user_token
        self._jwt = None
        self._urls = None

    @property
    def urls(self):
        if self._urls is None:
            if not self.base_url:
                raise NoDefaultUrl()
            self._urls = {
                'authorize': join(self.base_url, 'auth', 'authorize'),
                'account': join(self.base_url, 'account'),
                'account_tokens': join(self.base_url, 'account', 'tokens'),
                'login': join(self.base_url, 'auth', 'login'),
                'logout': join(self.base_url, 'logout'),
                'channels': join(self.base_url, 'channels'),
                'artifacts': join(self.base_url, 'artifacts'),
                'user_channels': join(self.base_url, 'account', 'channels'),
                'token_info': join(self.base_url, 'account', 'token-info'),
                'user_tokens': join(self.base_url, 'account', 'tokens'),
                'scopes': join(self.base_url, 'account', 'scopes'),
                'cves': join(self.base_url, 'cves'),
                'system_settings': join(self.base_url, 'system', 'settings'),
                'mirrors': join(self.base_url, "mirrors")
            }
        return self._urls

    def get_user_token_name(self):
        return 'repo-cli-token-{hostname}'.format(hostname=socket.gethostname())

    def create_access_token(self):
        # TODO: Should we remove expires_at and let the server pick the default?

        data = {
            'name': self.get_user_token_name(),
            'expires_at': str(datetime.datetime.now().date() + datetime.timedelta(days=30)),
            'scopes': self.get_scopes()
        }
        resp = requests.post(self.urls['account_tokens'], data=json.dumps(data), headers=self.bearer_headers, verify=self.verify_ssl)
        if resp.status_code != 200:
            msg = 'Error requesting a new user token! Server responded with %s: %s' % (resp.status_code, resp.content)
            logger.error(msg)
            raise errors.RepoCLIError(msg)

        self._access_token = resp.json()['token']
        return self._access_token

    @property
    def bearer_headers(self):
        headers = {
            'Authorization': f'Bearer {self._jwt}',
            'Content-Type': 'application/json'
        }
        return headers

    @property
    def is_admin_jwt(self):
        if not self._jwt:
            return False

        payload = self._jwt.split('.')[1]
        data = json.loads(base64.b64decode(payload + '=' * (len(payload) % 4)))

        roles = data['realm_access']['roles']
        return ROLE_ADMIN in roles

    @property
    def xauth_headers(self):
        return self.get_xauth_headers()

    def get_xauth_headers(self, extra_headers=None):
        headers = {'X-Auth': self._access_token}
        if extra_headers:
            headers.update(extra_headers)
        return headers

    def get_access_token(self):
        logger.debug('[API] Getting access token.. ')
        token_resp = requests.get(self.urls['account_tokens'], headers=self.bearer_headers, verify=self.verify_ssl)
        if token_resp.status_code != 200:
            msg = 'Error retrieving user token! Server responded with %s: %s' % (
            token_resp.status_code, token_resp.content)
            logger.error(msg)
            raise errors.RepoCLIError(msg)

        user_tokens = token_resp.json().get('items', [])
        logger.debug(f'[LOGIN] Access token retrieved.. {len(user_tokens)}')
        if user_tokens:
            user_token_name = self.get_user_token_name()
            user_tokens = [
                user_token
                for user_token in user_tokens
                if user_token['name'] == user_token_name
            ]
        if user_tokens:
            # ok, we got the token. Now we need to refresh it
            token_to_refresh = user_tokens[0]['id']
            refresh_url = join(self.urls['account_tokens'], token_to_refresh)
            logger.debug(f'[ACCESS_TOKEN] Refreshing token.. {token_to_refresh}')
            token_resp = requests.put(refresh_url, data="{}", headers=self.bearer_headers, verify=self.verify_ssl)
            if token_resp.status_code != 200:
                msg = 'Error refreshing user token! Server responded with %s: %s' % (
                token_resp.status_code, token_resp.content)
                logger.error(msg)
                raise errors.RepoCLIError(msg)
            self._access_token = new_token = token_resp.json()['token']
            logger.debug('[ACCESS_TOKEN] Token Refreshed')
            return new_token

        return None

    def get_authorize_url(self) -> 'URL':
        s = requests.Session()
        logger.debug(f'[LOGIN] Getting authorize endpoint')
        resp = s.get(self.urls['authorize'], verify=self.verify_ssl, allow_redirects=False)
        if resp.status_code != 302:
            logger.info('[LOGIN] Error getting auth configuration')
            logger.debug(f'Server responded with response {resp.status_code}')
            raise errors.WrongRepoAuthSetup()

        url = urlparse(resp.headers['Location'])
        url = url._replace(query='')
        return url

    def login(self, username, password) -> requests.Response:
        """ Login using direct grant and returns the jwt token. """
        data = {
            'username': username,
            'password': password
        }
        s = requests.Session()
        logger.debug(f'[LOGIN] Authenticating user {username}...')
        resp = s.post(self.urls['login'], data=json.dumps(data), headers={'Content-Type': 'application/json'}, verify=self.verify_ssl)
        logger.debug('[LOGIN] Done')

        if resp.status_code != 200:
            logger.info('[LOGIN] Error logging in...')
            logger.debug(f'Server responded with response {resp.status_code}\nData: {resp.content}')
            raise errors.Unauthorized()

        self._jwt = jwt_token = resp.json()['token']
        return jwt_token

    def get_user_token(self):
        """ Returns user token, used with X-Auth headers. """
        user_token = self.get_access_token()

        if not user_token:
            logger.debug('[LOGIN] Access token not found. Creating one...')
            # Looks like user doesn't have any valid token. Let's create a new one
            user_token = self.create_access_token()
            logger.debug('[LOGIN] Done.')

        # TODO: we are assuming the first token is the one we need... We need to improve this waaaaay more
        return user_token

    def get_default_channel(self):
        url = self.urls['account']
        logger.debug(f'[UPLOAD] Getting user default channel from {url}')
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, 'getting account details')['default_channel_name']

    def get_current_user(self):
        url = self.urls['account']
        logger.debug(f'[UPLOAD] Getting current user from {url}')
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, 'getting account details')

    def upload_file(self, filepath, channel, package_type):
        if package_type not in UPLOAD_TYPE_MAPPING:
            raise errors.RepoCLIError("%s upload is not supported" % package_type)

        artifact_type = UPLOAD_TYPE_MAPPING[package_type]

        url = join(self.base_url, 'channels', channel, 'artifacts')
        statinfo = os.stat(filepath)
        filename = basename(filepath)
        logger.debug(f'[UPLOAD] Using token {self._access_token} on {self.base_url}')
        multipart_form_data = {
            'content': (filename, open(filepath, 'rb')),
            'filetype': (None, artifact_type),
            'size': (None, statinfo.st_size)
        }
        logger.info(f'Uploading {package_type} artifact to {url}...')
        response = requests.post(url, files=multipart_form_data, headers=self.xauth_headers, verify=self.verify_ssl)
        return response

    def create_channel(self, channel):
        '''Create a new channel with name `channel` on the repo server at `base_url` using `token`
        to authenticate.

        Args:
              channel(str): name of the channel to be created

        Returns:
              response (http response object)
        '''
        _channel = channel
        logger.debug(f'Creating channel {_channel} on {self.base_url}')
        if '/' in channel:
            # this is a subchannel....
            channel, subchannel = channel.split('/')
            url = join(self.urls['channels'], channel, 'subchannels')
            data = {'name': subchannel}
            headers = self.get_xauth_headers({'Content-Type': 'application/json'})
        else:
            url = join(self.urls['channels'])
            data = {'name': channel}
            headers = self.get_xauth_headers({'Content-Type': 'application/json'})

        self._validate_channel_name(data['name'])
        response = requests.post(url, json=data, headers=headers, verify=self.verify_ssl)
        return self._manage_response(response, f'creating channel {_channel}', success_codes=[201])

    def remove_channel(self, channel):
        url = self._get_channel_url(channel)
        logger.debug(f'Removing channel {channel} on {self.base_url}')
        logger.debug(f'Using token {self._access_token}')
        response = requests.delete(url, headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        if response.status_code in [202]:
            logger.info(f'Channel {channel} successfully removed')
            logger.debug(f'Server responded with {response.status_code}\nData: {response.content}')
        else:
            msg = f'Error removing {channel}.' \
                f'Server responded with status code {response.status_code}.\n' \
                f'Error details: {response.content}'
            logger.error(msg)
            if response.status_code in [403, 401]:
                raise errors.Unauthorized()
        return response

    def update_channel(self, channel, success_message=None, **data):
        # we don't need to validate channel to change privacy
        if "privacy" not in data:
            self._validate_channel_name(channel)

        url = self._get_channel_url(channel)
        logger.debug(f'Updating channel {channel} on {self.base_url}')
        logger.debug(f'Using token {self._access_token}')
        response = requests.put(url, json=data, headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        if not success_message:
            success_message = f'Channel {channel} successfully updated.'
        if response.status_code in [200, 204]:
            logger.info(success_message)
            logger.debug(f'Server responded with {response.status_code}\nData: {response.content}')
        else:
            msg = f'Error creating {channel} .' \
                f'Server responded with status code {response.status_code}.\n' \
                f'Error details: {response.content}'
            logger.error(msg)
            if response.status_code in [403, 401]:
                raise errors.Unauthorized()
            # TODO: We should probably need to manage other error states
        return response


    def is_subchannel(self, channel):
        """Return True if channel is a path to a subchannel, False otherwise. For example:

        >> is_subchannel("main")
            False
        >> is_subchannel("main/stage")
            True

        Args:
            channel (str): name of the channel

        Returns:
            (bool)"""
        return '/' in channel

    def _get_channel_url(self, channel):
        """Return a channel url based on the fact that it's a normal channel or
         a subchannel

        Args:
            channel (str): name of the channel

        Returns:
            (str) url
        """
        if self.is_subchannel(channel):
            # this is a subchannel....
            channel, subchannel = channel.split('/')
            url = join(self.urls['channels'], channel, 'subchannels', subchannel)
        else:
            url = join(self.urls['channels'], channel)
        return url

    def get_channel(self, channel):
        logger.debug(f'Getting channel {channel} on {self.base_url}')
        url = self._get_channel_url(channel)
        response = requests.get(url, headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(response, f'getting channel {channel}')

    def get_channel_history(self, channel, offset=0, limit=50):
        logger.debug(f'Getting channel {channel} history {self.base_url}')
        url = join(self._get_channel_url(channel), 'history?offset=%s&limit=%s' % (offset, limit))
        response = requests.get(url, headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(response, f'getting channel {channel} history')

    def list_user_channels(self):
        logger.debug(f'Getting user channels from {self.base_url}')
        response = requests.get(self.urls['user_channels'],
                                headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(response, f'getting user channels')

    def list_channels(self):
        logger.debug(f'Getting channels from {self.base_url}')
        response = requests.get(self.urls['channels'],
                                headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(response, f'getting channels')

    def get_channel_subchannels(self, channel):
        logger.debug(f'Getting channel {channel} subchannels on {self.base_url}')
        url = join(self.urls['channels'], channel, 'subchannels')
        response = requests.get(url, headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(response, f'getting channel {channel} subchannel')

    def create_mirror(self, channel, source_root, name, mode, filters, type_, cron, run_now, proxy):
        url = join(self._get_channel_url(channel), 'mirrors')
        mirror_details = {
            "mirror_name": name,
            "source_root": source_root,
            "mirror_mode": mode,
            "cron": cron,
            "mirror_type": type_,
            "filters": filters,
            "run_now": run_now,
        }
        if proxy:
            mirror_details["proxy"] = proxy

        resp = requests.post(url, data=json.dumps(mirror_details), headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(resp, f'Creating mirror {name} on channel {channel}', success_codes=[201], auth_fail_codes=[401])

    def update_mirror(self, mirror_id, channel, source_root, name, mode, filters, type_, cron, run_now, proxy):
        url = join(self._get_channel_url(channel), 'mirrors', mirror_id)
        mirror_details = {
            "mirror_name": name,
            "source_root": source_root,
            "mirror_mode": mode,
            "cron": cron,
            "mirror_type": type_,
            "filters": filters,
            "run_now": run_now
        }
        if proxy:
            mirror_details["proxy"] = proxy

        resp = requests.put(url, data=json.dumps(mirror_details), headers=self.get_xauth_headers({'Content-Type': 'application/json'}), verify=self.verify_ssl)
        return self._manage_response(resp, f'Updating mirror {name} on channel {channel}', success_codes=[202], auth_fail_codes=[401])

    def get_mirrors(self, channel):
        url = join(self._get_channel_url(channel), 'mirrors')
        resp = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(resp, f'Getting mirrors on channel {channel}')

    def get_mirror(self, channel, mirror_name):
        url = join(self._get_channel_url(channel), 'mirrors', mirror_name)
        resp = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(resp, f'Getting mirror {mirror_name} on channel {channel}')

    def get_all_mirrors(self):
        url = self.urls["mirrors"]
        resp = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(resp, 'Getting global mirrors')

    def delete_mirror(self, channel, mirror_id, mirror_name):
        url = join(self._get_channel_url(channel), 'mirrors', mirror_id)
        resp = requests.delete(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(resp, f'Mirror {mirror_name} was deleted on channel {channel}',
                                     success_codes=[204])

    # TOKEN RELATED URLS
    def _manage_response(self, response, action='', success_codes=None, auth_fail_codes=None):
        if not success_codes:
            success_codes = [200]
        if not auth_fail_codes:
            auth_fail_codes = [401, 403]
        if response.status_code in success_codes:
            # deletes shouldn't return anythings
            if response.status_code == 204:
                return
            return response.json()
        else:
            msg = f'Error {action}. ' \
                  f'Server responded with status code {response.status_code}.\n' \
                  f'Error details: {response.content or None}\n'
            if response.status_code >= 500:
                msg += '\nPlease verify that the Anaconda Team Edition server is online and ' \
                       'that your configuration is pointing to the active ATE server.\n' \
                       'If Anaconda Team Edition is still responding with 500 errors, ' \
                       'please contact your system administrator.\n'
            if response.status_code == 401:
                logger.debug(msg)
            if response.status_code in auth_fail_codes:
                raise errors.Unauthorized()
            raise errors.RepoCLIError(msg)

    def get_token_info(self):
        response = requests.get(self.urls['token_info'], headers=self.xauth_headers, verify=self.verify_ssl)

        if response.status_code in [200, 204]:
            return response.json()
        else:
            msg = f'Error getting token info.' \
                  f'Server responded with status code {response.status_code}.\n' \
                  f'Error details: {response.content}'
            logger.error(msg)
            if response.status_code in [403, 401]:
                raise errors.Unauthorized()
        return {}

    def get_user_tokens(self):
        response = requests.get(self.urls['account_tokens'], headers=self.bearer_headers, verify=self.verify_ssl)

        if response.status_code in [200, 204]:
            return response.json()
        else:
            msg = f'Error getting user tokens.' \
                  f'Server responded with status code {response.status_code}.\n' \
                  f'Error details: {response.content}'
            logger.error(msg)
            if response.status_code in [403, 401]:
                raise errors.Unauthorized()
        return []

    def remove_user_token(self, token):
        url = join(self.urls['account_tokens'], token)
        response = requests.delete(url, headers=self.bearer_headers, verify=self.verify_ssl)
        return self._manage_response(response, 'removing user token', success_codes=[204])

    def create_user_token(self, name, expiration, scopes=None, resources=None):
        data = {
            'name': name,
            'expires_at': expiration,
        }
        if scopes:
            data['scopes'] = scopes

        if resources:
            data['resources'] = resources

        response = requests.post(self.urls['account_tokens'], data=json.dumps(data),
                                 headers=self.bearer_headers, verify=self.verify_ssl)
        return self._manage_response(response, "creating user token", success_codes=[200])

    def get_scopes(self):
        response = requests.get(self.urls['scopes'], headers=self.bearer_headers, verify=self.verify_ssl)
        return self._manage_response(response, "getting scopes")['scopes']

    # --------
    def channel_artifacts_bulk_actions(self, channel, action, artifacts, target_channel=None):
        url = join(self._get_channel_url(channel), 'artifacts', 'bulk')
        data = {
            "action": action,
            "items": artifacts
        }
        if target_channel:
            if '/' in target_channel:
                # this is a subchannel....
                data['target_channel'], data['target_subchannel'] = target_channel.split('/')
            else:
                data['target_channel'] = target_channel

        resp = requests.put(url, data=json.dumps(data),
                            headers=self.get_xauth_headers({'Content-Type': 'application/json'}),
                            # headers=self.bearer_headers
                            verify=self.verify_ssl
                            )
        return self._manage_response(resp, "%s articfacts" % action, success_codes=[202],
                                     auth_fail_codes=[401, 403, 404])

    def get_channel_artifacts(self, channel, offset=None, limit=None):
        query = []
        if offset is not None:
            query.append('offset=%s' % offset)
        if limit is not None:
            query.append('limit=%s' % limit)
        if query:
            url = join(self._get_channel_url(channel), 'artifacts?{}'.format('&'.join(query)))
        else:
            url = join(self._get_channel_url(channel), 'artifacts')
        resp = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(resp, "getting artifacts").get('items', [])

    def get_channel_artifacts_files(self, channel, family, package=None, version=None, filename=None):
        file_family_parsers = {
            'conda': self._parse_conda_file,
            'cran': self._parse_cran_file,
            'python': self._parse_python_file,
            'anaconda_project': self._parse_project_file,
            'anaconda_env': self._parse_environment_file,
            'notebook': self._parse_notebook_file
        }
        artifact_files = []
        total_count = 0
        if package:
            packages = [package]
        else:
            # url = join(self.urls['channels'], channel, 'artifacts')
            # resp = requests.get(url, headers=self.xauth_headers)
            # data = self._manage_response(resp, "getting articfacts")
            data = self.get_channel_artifacts(channel)
            packages = [pkg['name'] for pkg in data if pkg['file_count'] > 0 and pkg['family'] == family]

        for package in packages:
            url = join(self._get_channel_url(channel), 'artifacts', family, package, 'files')
            resp = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
            files = self._manage_response(resp, "getting articfacts")
            total_count = files["total_count"]
            for file_data in files['items']:
                rec = file_family_parsers[family](file_data, version, filename,)
                if rec:
                    for cve_field in ['cve_status', 'cve_score']:
                        if cve_field in file_data:
                            rec[cve_field] = file_data[cve_field]
                    artifact_files.append(rec)

        return artifact_files, total_count

    def _parse_conda_file(self, file_data, version, filename, return_raw = False):
        meta = file_data["metadata"].get("repodata_record.json") or file_data['metadata']['index.json']
        # TODO: We need to improve version checking... for now it's exact match
        if version and meta['version'] != version:
            return
        if filename and meta['fn'] != filename:
            return
        if return_raw:
            rec = file_data
        else:
            rec = {'name': file_data['name'], 'ckey': file_data['ckey']}
            rec.update({key: meta.get(key, "") for key in ['version', 'fn']})
            rec['platform'] = meta.get('subdir', '')

        return rec

    def _parse_cran_file(self, file_data, version, filename, return_raw=False):
        meta = file_data['metadata']
        fn = file_data['ckey'].split('/')[-1]

        # TODO: We need to improve version checking... for now it's exact match
        if version and meta['Version'] != version:
            return
        if filename and fn != filename:
            return
        if return_raw:
            rec = file_data
        else:
            rec = {'name': file_data['name'], 'ckey': file_data['ckey']}
            # rec.update({key: meta.get(key, "") for key in ['version', 'fn', 'platform']})
            rec.update(meta)
            rec['version'] = rec.pop('Version')
            rec['fn'] = fn
            rec['platform'] = 'n/a'
        return rec

    def _parse_project_file(self, file_data, version, filename, return_raw = False):
        fn = file_data['ckey'].split('/', maxsplit=1)[-1]
        platforms = file_data['metadata'].get('platforms')
        if isinstance(platforms, list):
            platforms = ','.join(platforms)
        rec = {
            'name': file_data['name'],
            'ckey': file_data['ckey'],
            'fn': fn,
            'platform': platforms if platforms else 'n/a',
            'version': None
        }
        return rec

    def _parse_environment_file(self, file_data, version, filename, return_raw = False):
        fn = file_data['ckey'].split('/', maxsplit=1)[-1]
        rec = {
            'name': file_data['name'],
            'ckey': file_data['ckey'],
            'fn': fn,
            'platform': 'n/a',
            'version': None
        }
        return rec

    def _parse_notebook_file(self, file_data, version, filename, return_raw = False):
        fn = file_data['ckey'].split('/', maxsplit=1)[-1]
        rec = {
            'name': file_data['name'],
            'ckey': file_data['ckey'],
            'fn': fn,
            'platform': 'n/a',
            'version': None
        }
        return rec

    def _parse_python_file(self, file_data, version, filename, return_raw=False):
        meta = file_data['metadata']
        artifact_filename = file_data['ckey'].split('/')[-1]

        if version and meta['version'] != version:
            return

        if filename and artifact_filename != filename:
            return

        if return_raw:
            rec = file_data
        else:
            rec = {'name': file_data['name'], 'ckey': file_data['ckey']}
            rec.update(meta)
            rec['version'] = rec.pop('version')
            rec['artifact_filename'] = artifact_filename
            rec['platform'] = 'n/a'

        return rec

    def get_artifacts(self, query, limit=50, offset=0, sort="-download_count"):
        url = "%s?q=%s&limit=%s&offset=%s&sort=%s" %(self.urls['artifacts'], query, limit, offset, sort)
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, "searching artifacts")

    # CVE related endpoints
    def get_cves(self, offset=0, limit=50):
        url = "%s?offset=%s&limit=%s" % (self.urls['cves'], offset, limit)
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, "getting cves")

    def get_cve(self, cve_id):
        url = "%s/%s" % (self.urls['cves'], cve_id)
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, "getting cve id")

    def get_cve_files(self, cve_id, offset=0, limit=50):
        url = "%s/%s/files?offset=%s&limit=%s" % (self.urls['cves'], cve_id, offset, limit)
        response = requests.get(url, headers=self.xauth_headers, verify=self.verify_ssl)
        return self._manage_response(response, "getting cve files")

    def _validate_channel_name(self, name: str):
        if not re.match(r"^[a-z][a-z0-9_-]*$", name):
            # try to assist user by pointing to invalid letters
            invalid_chars = set(r"""!"#$%&'()*+,./:;<=>?@[\]^`{|}~""")
            invalid_letters = list(invalid_chars.intersection(set(name)))
            error_message = f"Channel name contains invalid sequence {', '.join(invalid_letters)}"
            logger.error(error_message)
            raise InvalidName(error_message)

    # Admin settings endpoints
    def get_system_settings(self):
        response = requests.get(
            self.urls["system_settings"],
            headers=self.xauth_headers
        )
        return self._manage_response(response, "getting system settings")

    def update_system_settings(self, settings):
        response = requests.put(
            self.urls["system_settings"],
            data=json.dumps(settings),
            headers=self.get_xauth_headers({'Content-Type': 'application/json'}),
            verify=self.verify_ssl
        )
        return self._manage_response(response, "updating system settings")



"""
Manage your Anaconda repository channels.
"""

from __future__ import unicode_literals, print_function

import argparse
from pprint import pformat

from .base import SubCommandBase
from .. import errors
from ..utils import format
from ..utils.artifacts import SimplePackageSpec


class SubCommand(SubCommandBase):
    name = "channel"

    def main(self):
        self.log.info("")
        if self.args.create:
            channel = self.args.create
            if not channel:
                msg = 'Channel name not specified. Please use -n or -o to specify your channel.\n' \
                      'Use --help for help.'
                self.log.info(msg)
                raise errors.RepoCLIError(msg)

            self.api.create_channel(channel)
            self.log.info("Channel %s successfully created" % channel)
        elif self.args.remove:
            self.api.remove_channel(self.args.remove)
        elif self.args.list:
            self.list_user_channels()
        elif self.args.list_packages:
            for channel_name in self.args.list_packages:
                self.show_channel_packages(channel_name, self.args.offset, self.args.limit, self.args.full_details)
        elif self.args.list_files:
            for spec in self.args.list_files:
                self.show_channel_files(spec, family=self.args.family, full_details=self.args.full_details)
        elif self.args.show:
            self.show_channel(self.args.show, full_details=self.args.full_details)
        elif self.args.history:
            self.show_channel_history(self.args.history, self.args.offset, self.args.limit, self.args.full_details)
        elif self.args.lock:
            channel = self.args.lock
            msg = "{} {} is now locked".format(self.name.title(), channel)
            self.api.update_channel(channel, privacy='private', success_message=msg)
        elif self.args.soft_lock:
            channel = self.args.soft_lock
            msg = "{} {} is now soft-locked".format(self.name.title(), channel)
            self.api.update_channel(channel, privacy='authenticated', success_message=msg)
        elif self.args.unlock:
            channel = self.args.unlock
            msg = "{} {} is now unlocked".format(self.name.title(), channel)
            self.api.update_channel(channel, privacy='public', success_message=msg)
        else:
            raise NotImplementedError()

    def filter_response(self, data, offset=None, limit=None):
        filtered_data = data[offset:(limit + offset if limit is not None else None)]
        return filtered_data

    def show_channel_packages(self, channel_name, offset, limit, full_details):
        packages = self.api.get_channel_artifacts(channel_name, offset, limit)

        self.log.info('')
        self.log.info('Total packages matching spec %s found: %s\n' % (channel_name, len(packages)))

        if full_details:
            for package in packages:
                self.show_package_detail(package)
            self.log.info('')
        else:
            formatter = format.PackagesFormatter(log=self.log)
            self.log.info(formatter.format(packages=packages, metadata={}, simple=True))

    def show_package_detail(self, package):
        keymap = {'download_count': '# of downloads', 'file_count': '# of files',}
        pack = dict(package)
        pack.update(package['metadata'])
        resp = ["---------------"]

        for key in ['name', 'file_count', 'download_count', 'license', 'description']:
            label = keymap.get(key, key)
            value = pack.get(key, '')
            resp.append("%s: %s" % (label, value))

        self.log.info('\n'.join(resp))

    def show_channel_files(self, spec, family, full_details=False):
        packages, total_count = self.api.get_channel_artifacts_files(
            spec.channel, family, spec.package, spec.version, spec.filename
        )
        if not packages:
            self.log.warning('No files matches were found for the provided spec: %s\n' % (spec))
            return

        formatter = format.PackageFilesFormatter
        if 'cve_score' in packages[0]:
            formatter = format.PackageFilesFormatterWithCVE

        packages = self.filter_response(packages, self.args.offset, self.args.limit)
        total_count = len(packages)
        self.log.info('Found %s files matching the specified spec %s:' % (total_count, spec))

        if full_details:
            self._show_channel_files_full_format(packages)
        else:
            self.log.info('')
            self.log.info(formatter.format_list(packages))
            self.log.info('')

    def _show_channel_files_full_format(self, packages):
        for package in packages:
            self.log.info('----------------')
            self.log.info(pformat(package))

    def show_channel(self, channel, full_details=False):
        channel_data = self.api.get_channel(channel)
        if full_details and not self.api.is_subchannel(channel):
            channel_data['subchannels'] = self.api.get_channel_subchannels(channel)

        self.show_channel_detail(channel_data)

    def show_channel_history(self, channel, offset, limit, full_details):
        data = self.api.get_channel_history(channel, offset, limit)
        self.log.info(format.HistoryFormatter.format_list(data['items'], not full_details))
        self.log.info('')

    def show_channel_detail(self, data):
        resp = ["Channel details:", '']
        keymap = {'download_count': '# of downloads', 'artifact_count': '# of artifacts',
                  'mirror_count': '# mirrors', 'subchannel_count': '# of subchannels'}

        for key in ['name', 'description', 'privacy', 'artifact_count', 'download_count', 'mirror_count',
                    'subchannel_count', 'created', 'updated']:
            label = keymap.get(key, key)
            value = data.get(key, '')
            resp.append("\t%s: %s" % (label, value))

        try:
            owners = ', '.join(data.get('owners', ['']))
        except TypeError:
            owners = ''
        resp.append('\towners: %s' % owners)
        if 'subchannels' in data:
            resp.append("")
            resp.append("\tSubchannels:")
            resp.append("\t------------")

            resp.append(format.PackagesFormatter.format_channel_header())
            for subchannel in data['subchannels']['items']:
                resp.append(format.PackagesFormatter.format_channel(subchannel))

        resp.append("")
        self.log.info('\n'.join(resp))

    def list_user_channels(self):
        self.show_list_channels(self.api.list_user_channels())

    def show_list_channels(self, data):
        resp = ["Channels available to the user:", '']
        keymap = {'download_count': 'downloads', 'artifact_count': 'artifacts'}
        cols_ = ['name', 'privacy', 'description', 'artifact_count', 'download_count']
        cols = [keymap.get(key, key) for key in cols_]
        resp.append('\t'.join(cols))

        for ch in data['items']:
            resp.append('\t'.join([str(ch.get(key, '')) for key in cols_]))

        resp.append('')
        self.log.info('\n'.join(resp))

    def add_parser(self, subparsers):
        subparser = subparsers.add_parser(
            self.name,
            help='Manage your Anaconda repository {}s'.format(self.name),
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=__doc__
        )
        subparser.add_argument(
            '--family',
            default='conda',
            help='Artifact family (i.e.: conda, python, cran, anaconda_project, anaconda_env, notebook). '
                 'ONLY USED IN COMBINATION WITH --list-files, ignored otherwise.'
        )
        subparser.add_argument(
            '--full-details',
            help='Prints full file details. ONLY USED IN COMBINATION WITH --list-files, --list-packages '
                 'or --show ignored otherwise.',
            action='store_true',
            default=False
        )
        subparser.add_argument(
            '-o', '--offset',
            default=0,
            type=int,
            help='Offset when displaying the results'
        )
        subparser.add_argument(
            '-l', '--limit',
            default=50,
            type=int,
            help='Limit when displaying the results',
        )

        group = subparser.add_mutually_exclusive_group(required=True)

        group.add_argument(
            '--create',
            # action='store_true',
            metavar=self.name.upper(),
            help="Create a new {}".format(self.name)
        )
        group.add_argument(
            '--list',
            action='store_true',
            help="List all {}s for a user".format(self.name)
        )
        group.add_argument(
            '--list-packages',
            help='Channel written as <channel>[/<subchannel>]',
            metavar='CHANNEL',
            nargs='+'
        )
        group.add_argument(
            '--list-files',
            help='Package written as <channel>[/<subchannel>]::<package>[/<version>[/<filename>]]',
            metavar='PACKAGE',
            type=SimplePackageSpec.from_string,
            nargs='+'
        )
        group.add_argument(
            '--show',
            metavar=self.name.upper(),
            help="Show information about the {}".format(self.name)
        )
        group.add_argument(
            '--history',
            metavar=self.name.upper(),
            help="Show historical events about the {}".format(self.name)
        )
        group.add_argument(
            '--lock',
            metavar=self.name.upper(),
            help="Lock a {}".format(self.name)
        )
        group.add_argument(
            '--soft-lock',
            metavar=self.name.upper(),
            help="Soft Lock a {}, so that only authenticated users can see it.".format(self.name)
        )
        group.add_argument(
            '--unlock',
            metavar=self.name.upper(),
            help="Unlock a {}".format(self.name)
        )
        group.add_argument(
            '--remove',
            metavar=self.name.upper(),
            help="Remove a {}".format(self.name)
        )

        subparser.set_defaults(main=self.main)
